
"""

Graphical operations on video frames.

We encode and decode videos by forking an mencoder process and
streaming raw video frames in RGB to its standard input and
output. This may not be very fast, but it is simpler than linking
with video libraries.


"""

import os,sys,select,time,random,traceback,popen2

# clear to end of line, nice in combination with \r
clr_eol="\033[K"


##########################################################################################
# I/O on video files
##########################################################################################


class VidPlayer:
  def __init__(self):
    self.t=-1
    self.expected_tend=None
    
  def get_frame(self):
    """ returns a frame as raw RGB data """
    self.t+=1
    return None


class FileVidPlayer(VidPlayer):

  def get_frame(self):
    """ read the required bytes from file descriptor f """
    self.t+=1

    data=""
    while len(data)<self.datasize:
      r=self.f.read(self.datasize-len(data))
      if not r: break
      data+=r      

    if len(data) not in (0,self.datasize): 
      print "WARN: incomplete frame (%d unused bytes)"%len(data)      
      data=None
    if not data:
      return None
    
    # assert len(data)==self.datasize,"got %s/%d"%(repr(data),self.datasize)
    
    return data

  def __del__(self):
    """ need special care because the mencoder process does not
    necessarily exit gracefully on sigpipe"""

    if self.child_pid!=None:
      os.kill(self.child_pid,9)
      if self.dec_pipe: 
        os.unlink(self.dec_pipe)
    else:
      pass


class FlexPlayer(FileVidPlayer):

  def __init__(self,fname,fmt="rgb24",skip_secs=0.0,verbose_mplayer=False,add_vf=None):
    VidPlayer.__init__(self)
    self.child_pid=None
    self.dec_pipe=None

       
    # first find the input video size

    self.w=self.h=self.fps=self.video_length=None
        
    cmd="mplayer -identify -vo null -ao null -frames 0 \"%s\" %s"%(
      fname,verbose_mplayer and " " or "2> /dev/null")

    for line in os.popen(cmd,"r"):
      fields=line.split("=")
      if fields[0]=="ID_VIDEO_WIDTH": self.w=int(fields[1])
      if fields[0]=="ID_VIDEO_HEIGHT": self.h=int(fields[1])
      if fields[0]=="ID_VIDEO_FPS": self.fps=float(fields[1])
      if fields[0]=="ID_LENGTH": self.video_length=float(fields[1])
      
    assert self.w and self.h and self.fps,"something wrong, does file '%s' exist? is it a video file?"%fname

    print 'w = %d, h = %d, fps = %.3f, length = %.3f\n' % (self.w, self.h, self.fps, self.video_length)

    vf="format="+fmt

    if add_vf:
      vf+=","+add_vf
            
    if fmt=="rgb24":
      self.datasize=self.w*self.h*3
    elif fmt=="YV12":
      self.datasize=self.w*self.h+self.w*self.h/2
    else:
      assert 0

    self.dec_pipe=os.tempnam()

    os.mkfifo(self.dec_pipe)

    pid=os.fork()

    if pid==0: 

      args=["mencoder",fname,
            "-of","rawvideo",
            "-oac","pcm","-af","export=/dev/zero",
            # "-nosound",
            "-ovc","raw","-vf",vf,
            "-o",self.dec_pipe,"-ss","%g"%skip_secs]
 
      if not verbose_mplayer:
        f=open("/dev/null","w")
        os.dup2(f.fileno(),1)
        # os.dup2(f.fileno(),2)
        args+=["-quiet","-msglevel","all=0"]
       
      os.execvp("mencoder",args)

    self.f=open(self.dec_pipe,"r")
    self.child_pid=pid




class VidEncoder:

  def __init__(self,fname,w,h,fps,verbose_mplayer=False):
    self.w=w
    self.h=h

    valid_rates=[15,25,30,60] # known to work with mpeg-1
    if fps not in valid_rates:
      print "framerate %g not valid for MPEG-1, using"%fps,
      fps=min([(abs(r-fps),r) for r in valid_rates])[1]
      print fps
      
    self.fps=fps     
    
    self.child_pid=None
    self.enc_pipe=None    
    self.enc_pipe=os.tempnam()

    os.mkfifo(self.enc_pipe)
    
    pid=os.fork()

    if pid==0: 

      args=["mencoder",
            "-nosound",
            # input params            
            "-demuxer","rawvideo","-rawvideo",
            "w=%d:h=%d:format=rgb24:fps=%g"%(self.w,self.h,self.fps),
            self.enc_pipe,
            # output
            # "-of","lavf","-ovc","x264","-x264encopts","bitrate=1000", # generates nonconformant .mp4 files
            "-ovc","lavc","-lavcopts","vcodec=mpeg1video:vbitrate=1150000:keyint=12",
            "-o",fname] 

      if not verbose_mplayer:
        f=open("/dev/null","w")
        os.dup2(f.fileno(),1)
        os.dup2(f.fileno(),2)
        args+=["-quiet","-msglevel","all=0"]
             
      os.execvp("mencoder",args)

    self.f=open(self.enc_pipe,"w")
    self.child_pid=pid

  def add_frame(self,data):
    assert len(data)==self.w*self.h*3
    self.f.write(data)

  def __del__(self):
    """ need special care because the mencoder process does not
    necessarily exit gracefully on sigpipe"""

    if self.child_pid!=None:
      os.kill(self.child_pid,9)
    if self.enc_pipe: 
      os.unlink(self.enc_pipe)





##########################################################################################
# Simple Gimp-based image filters
##########################################################################################

    

from gimpfu import *


class SimpleFilter:
  " filter depending on a single parameter "
  
  def __init__(self,param=0):
    self.param=param

  def filter(self,im,lay):
    " transforms frame stored as a GIMP layer "
    pass


class NoiseFilter(SimpleFilter):
  def filter(self,im,lay):
    pdb.plug_in_rgb_noise(im,lay,True,False,self.param,self.param,self.param,0)

class GammaFilter(SimpleFilter):
  def filter(self,im,lay):
    pdb.gimp_curves_spline(lay, 0, 6, [0,0,128,255*0.5**self.param,255,255])
  
class ContrastFilter(SimpleFilter):
  def filter(self,im,lay):
    pdb.gimp_brightness_contrast(lay,0,self.param)
  
class BlurFilter(SimpleFilter):
  def filter(self,im,lay):
    pdb.plug_in_gauss(im,lay, self.param, self.param, 1)
    
class PixelizeFilter(SimpleFilter):
  def filter(self,im,lay):
    pdb.plug_in_pixelize(im, lay, self.param)

class ScaleUnscaleFilter(SimpleFilter):
  " blurs frames by scaling them down and scaling up again "
  
  def filter(self,im,lay):
    lay2=pdb.gimp_layer_copy(lay, True)
    pdb.gimp_image_add_layer(im,lay2, -1)
    lay2 = pdb.gimp_drawable_transform_scale_default(
      lay2, 0, 0, im.width/self.param, im.height/self.param, True, 0)
    # nice crappy scale method :-)
    lay2=pdb.gimp_drawable_transform_scale(
      lay2, 0, 0, im.width, im.height, 0, 1, True, 1, 0)

    pdb.gimp_edit_copy(lay2)
    pdb.gimp_image_remove_layer(im,lay2)

    floating_sel=pdb.gimp_edit_paste(lay, False)
    pdb.gimp_floating_sel_anchor(floating_sel)    

    



##########################################################################################
# Complex Gimp-based image filters
##########################################################################################


  
class GainControl:
  """ Mimic automatic gain ctrl on video camera.
  strength: how strong is the adjustement (0=no adjustment, 1=violent adjustment)
  inertia: how fast the camera adjusts to changes in image brihghtness (0=immediately, 1=never)  
  """
  
  def __init__(self,strength=0.8,inertia=0.8):
    self.strength=strength
    self.inertia=inertia
    self.prev=None
        
  def filter(self,im,lay):

    mean, std_dev, median, pixels, count, percentile = pdb.gimp_histogram(
      lay, HISTOGRAM_VALUE, 0, 256)

    if self.prev:
      # mix with previous stats
      prev_mean,prev_std_dev=self.prev
      mean=prev_mean*self.inertia+mean*(1-self.inertia)
      std_dev=prev_std_dev*self.inertia+std_dev*(1-self.inertia)

    self.prev=mean,std_dev    
    
    fact=0.5

    # this value will be mapped to value 10 
    map_to_10 = mean - std_dev * fact
    if map_to_10<15: map_to_10=15
    
    # this value will be mapped to value 255 
    map_to_255 = mean + std_dev * fact
    if map_to_255>240: map_to_255=240
    
    if map_to_10+10>map_to_255:
      avg=0.5*(map_to_10+map_to_255)
      map_to_10,map_to_255=avg-5,avg+5

    # apply stength
    map_to_10=self.strength*map_to_10+(1-self.strength)*10
    map_to_255=self.strength*map_to_255+(1-self.strength)*255
    
    curve=[0,0,                                     
           map_to_10,10,
           map_to_255,255,
           255,255]

    pdb.gimp_curves_spline(lay, 0, 8, curve)


  
class PerspectiveFilter:
  """
  Perspective transform on the input image.
  coords=table of 4*2 corner coordinates of the input image in the output image   
  """

  def __init__(self,coords):
    self.coords=coords   
  
  def filter(self,im,lay):
    lay2=pdb.gimp_layer_copy(lay, True)

    pdb.gimp_context_set_background((0,0,0))
    pdb.gimp_drawable_fill(lay, BACKGROUND_FILL)

    pdb.gimp_image_add_layer(im,lay2, -1)

    coords=self.coords

    lay2=pdb.gimp_drawable_transform_perspective_default(
      lay2,
      coords[0],coords[1],coords[2],coords[3],
      coords[4],coords[5],coords[6],coords[7],
      True, 1)

    pdb.gimp_edit_copy(lay2)

    pdb.gimp_image_remove_layer(im,lay2)
        
    floating_sel=pdb.gimp_edit_paste(lay, False)
    pdb.gimp_floating_sel_anchor(floating_sel)        

    
    

class Shake2D:
  """ Brownian motion to simulate shake
  The motion is made of time segments where the coordinates move along a line.
  """  
  def __init__(self,rmag,tmag):
    self.rmag=rmag
    self.tmag=tmag      

    # current segment time bounds
    self.t0=-1.0
    self.t1=0.0
    # position at t0
    self.p_x=self.p_y=0.0
    # vector to t1
    self.v_x=self.v_y=0.0
    self.t=0.0
    # current coordinates
    self.coords=(0.0,0.0)

  def next_seg(self):
    " generate a new segment "
    self.t0=self.t1
    self.p_x+=self.v_x
    self.p_y+=self.v_y
    
    # shorter segments are more likely
    self.t1+=random.expovariate(1.0/self.tmag)

    # not completely random: the coordinates tend to return to the center
    v=math.hypot(self.p_x,self.p_y)
    if v>0:  cx,cy=-self.p_x/v*self.rmag,-self.p_y/v*self.rmag
    else:    cx,cy=0,0
    
    self.v_x=random.gauss(cx,self.rmag)
    self.v_y=random.gauss(cy,self.rmag)

  def seg_coords(self,t):
    t=(t-self.t0)/(self.t1-self.t0)
    u=math.sin(t*math.pi)**2
    return (self.p_x+u*self.v_x,
            self.p_y+u*self.v_y)

  def step(self,dt=1.0):
    " step forward by dt and update current coordinates "
    assert dt>0
    self.t+=dt
    while self.t>self.t1:
      self.next_seg()
    self.coords=self.seg_coords(self.t)
    
class PerspectiveWithShakeFilter(PerspectiveFilter):
  
  def __init__(self,coords,shake):
    PerspectiveFilter.__init__(self,coords)
    self.shake=shake
    
  def filter(self,im,lay):

    coords=self.coords
    
    shutter=0.9

    # position when camera shutter opens
    shake=self.shake
    shake.step(1.0-shutter)
    (dx1,dy1)=shake.coords

    # position when camera shutter closes
    shake.step(shutter)
    (dx2,dy2)=shake.coords
    npix=math.hypot(dx1-dx2,dy1-dy2)
    ncopy=int(math.ceil(npix+0.0001))  

    # draw the perspective-distorted frame several times, to simulate
    # motion blur due to shaking
    
    accu=pdb.gimp_layer_copy(lay, True)   # accumulation layer
    pdb.gimp_image_add_layer(im,accu, -1)
        
    for copy in range(ncopy):
      
      frac=(copy+1)/float(ncopy)
      (dx,dy)=(dx1*(1-frac)+dx2*frac,
               dy1*(1-frac)+dy2*frac)

      lay2=pdb.gimp_layer_copy(lay, True)
      pdb.gimp_image_add_layer(im,lay2, -1)
      
      lay2=pdb.gimp_drawable_transform_perspective_default(
        lay2,
        coords[0]+dx,coords[1]+dy,coords[2]+dx,coords[3]+dy,
        coords[4]+dx,coords[5]+dy,coords[6]+dx,coords[7]+dy,
        True, 1)

      # trick to make layer opaque
      curve=[0,255,255,255]
      pdb.gimp_curves_spline(lay2,4,4,curve)

      lay2.opacity=100.0/(1+copy)

      # accumulate 
      accu=pdb.gimp_image_merge_down(im,lay2,1)

    # copy accumulator to initial layer
    pdb.gimp_edit_copy(accu)        
    pdb.gimp_image_remove_layer(im,accu)
    
    floating_sel=pdb.gimp_edit_paste(lay, False)
    pdb.gimp_floating_sel_anchor(floating_sel)        



##########################################################################################
# Run a filter chain
##########################################################################################



def run_filter(stage1src,filters,sink):
  """
  stage1src should be a VidPlayer
  filters is a list of objects with a filter(im,layer) method
  sink is a VidSink
  """ 
  
  w,h=stage1src.w,stage1src.h
  im=pdb.gimp_image_new(w,h,0)
  lay=pdb.gimp_layer_new(im,w,h,0,"i/o layer",100, 0)

  pdb.gimp_image_add_layer(im,lay,0)

  pr=lay.get_pixel_rgn(0,0,im.width,im.height,True,False)

  print "Begin transform"

  t=0
  
  while True:
    data=stage1src.get_frame()

    if not data: break

    print clr_eol,"frame",t,"\r",
    sys.stdout.flush()
    
    pr[0:w,0:h]=data

    for filt in filters:
      filt.filter(im,lay)

    # f=open("data/frames/%03d.ppm"%t,"w")
    # f.write("P6\n%d %d\n255\n"%(w,h))
    # f.write(pr[0:w,0:h])

    sink.add_frame(pr[0:w,0:h])

    t+=1

  pdb.gimp_image_delete(im)
  print "End transform"

